c  H Y P E R N D - simplified N-dimensional hyperdynamics
c
c  Current (6/2001) potentials available are:
c    * model potential from JCP 106, 4665 (1997) with optional slope
c        (for biased diffusion) or additional y-like degrees of freedom
c    * Frenkel-Kontorova chain
c
c  The JCP (1997) analytic bias potential (Eq. 31) is available for
c    the 2-D model potential; flat bias can be used for any potential
c
c  Written by Art Voter, 1/2000; Modified by Tim Germann, 5/2001
c  This is a work in progress - please do not distribute or copy without 
c  permission from Art Voter.

      implicit real*8(a-h,o-z)
      parameter(noutermax=50000)
      parameter(maxtau=50000)
      dimension crosscount(noutermax)
      dimension tauobserved(maxtau)
      dimension temps(noutermax)

      parameter (nmax=100)  ! degrees of freedom
      dimension x(nmax) 
      dimension velx(nmax),dvdx(nmax),accelx(nmax)
      dimension accelnewx(nmax)
      dimension randx(nmax),randvx(nmax)
      character*80 title,fmt

c-------------------------------------------------------
c Some notes on the use of this hypernd program.

c  1) Try the sample input below.  It runs in less than a minute.

c  2) The hyperdynamics method is presented in the following
c     two papers:

c        A.F. Voter, J. Chem. Phys. 106, 4665 (1997),
c     and
c        A.F. Voter, Phys. Rev. Lett. 78, 3908 (1997).

c  3) iprint=1 generates output on the following units:
c        unit 9 - time, energy, etc, after every traj section
c        unit 11 - record of every transition
c     iprint=2 gives some output with every MD step (unit 13, big!)

c  4) You may want to tailor the output to suit your needs.

c  5) There is a typo in the potential in the JCP paper - 
c     see hyperndpots.f for the correct form.

c  6) The flat bias method implemented here is based on the
c     proposal of Steiner et al:

c        M. M. Steiner, P.-A. Genilloud, and J. W. Wilkins,
c        Phys. Rev. B   57, 10236 (1998).
c
c    except that I am currently doing nothing about smoothing
c    the cusp where the flat bias potential meets the real 
c    potential.  For now, I am simply using a shorter time
c    step and checking that the average temperature and the
c    variance of the kinetic temperature are accurate.

c  7) All coordinates have the same mass (unity) for now.
c-------------------------------------------------------


c-------------------------------------------------------
c Sample data for a test run on Voter JCP model potential I.
c For this potential (icase=1), minimum is -1.203 and saddle is +0.797
c
c The crossing count can be compared with  the kTST entry in Table II
c in the JCP paper (A.F. Voter, J. Chem. Phys. 106, 4665 (1997).
c (Note that in that paper a different type of bias potential 
c  was employed.  Also, to get the real diffusion constant requires
c  looking at the slope of the mean-square displacement curve.)

c Decomment the following lines for a test run:
c Sample run, JCP potential I, kT=0.2, vflat=0.6
c 1     ! icase (potential number - see potset in hyperndpots.f)
c 2     ! n (number of degrees of freedom)
c 0.01  ! dt (time step - small for cusps in bias pot)
c 100   ! nstep  (steps per section of trajectory)
c 5000  ! nouter (number of trajectory sections)
c 0.5   ! beta (friction coefficient)
c 0.2   ! tk  (Boltzmann*temperature)
c 1.5   ! zmax (bias potential parameter; negative -> use flat bias)
c 0.004 ! a (bias potential parameter) or vflat if zmax < 0
c 1     ! iprint (print level)
c-------------------------------------------------------

      write(6,*) '  HYPERND program -- afv, 1/19/2000;  tcg, 6/3/2001'
      write(6,*) ' '

c read in the data
      read(5,'(a80)') title
      read(5,*) icase
      read(5,*) n
      read(5,*) dt
      read(5,*) nstep
      read(5,*) nouter
      read(5,*) beta
      read(5,*) tk
      read(5,*) zmax
      read(5,*) a
      read(5,*) iprint
      
      i1=0
      call seedit(i1)

      write(6,*) title
      write(6,70) icase, n
 70   format('icase =',i2, ', n =',i3, ' degrees of freedom')
      write(6,72) dt, nstep, nouter
 72   format('dt =',f6.3, ', nstep =',i6, ', nouter =',i6)
      write(6,74) beta, tk
 74   format('beta =',f7.4, ', tk =',f7.4)
      if (zmax .lt. 0.d0) then
         vflat = a
         write(6,76) vflat
 76      format('vflat =',f7.4,
     &        ' (flat bias potential, a la Steiner et al)')
      else
         write(6,78) zmax, a
 78      format('zmax =',f8.4, ', a =', f9.5,
     &        ' (bias potential Eq. (31) of Voter JCP 1997)')
      end if
      write(6,80) iprint
 80   format('iprint =',i2)

c set up the potential
      call potset(n,icase)
      if (icase.le. 2) call mep_jcp(n, zmax, a, vflat)

c some other system-dependent parameters
      rmass=1.0d0

      tmd=0.0d0
      thyper=0.0d0

      do i=1,n
         x(i)=0.5d0 
         velx(i)=0.2d0*(-1.0d0)**i
      end do
      istatelast=kstate(x,n)

      thyperlast=0.0d0
      ncrosscum=0

      if (iprint.ge.1) write(9,90) 
   90 format('       tmd        thyper          tkkin        V',         
     x  '          V+delV       boostfac        T+V     ',
     x  '       x           x2')

c set up coeficients for Brownian integration (Allen and Tildesley, p. 261)
      bdt=beta*dt
      c0=exp(-bdt)              ! roughly 1 - bdt 
      c1=(1.0d0-c0)/bdt         ! roughly 1 - bdt/2
      c2=(1.0d0-c1)/bdt         ! roughly 1/2 - bdt/6

c set up variances and correlation coef for x and v displacements (A&T, p. 262)
      bdt=beta*dt
      sig2r=tk/rmass/beta/beta *
     x     (2.d0*bdt-3.d0+4.d0*exp(-bdt)-exp(-2.d0*bdt))
      sig2v=tk/rmass*(1.0d0-exp(-2.d0*bdt))
      sigrsigvcrv=tk/rmass/beta*(1.0d0-exp(-bdt))**2
      sigr=sqrt(sig2r)
      sigv=sqrt(sig2v)
      crv=sigrsigvcrv/(sigr*sigv)

      write(6,*) ' '
      write(6,*) ' running trajectory ...'
      write(6,*) ' '

      if(iprint.ge.1) then
       write(11,*)'ncrosscum,tmd,thyper,thyper-thyperlast,kstate,iphase'
       write(11,95) ncrosscum,tmd,thyper,thyper-thyperlast,kstate(x,n),0
      end if

      call v(x,n, vnow)
      if (zmax .lt. 0.d0) then
         delv = vbias_flat(x, n, vnow, dvdx, vflat)
      else
         delv = vbias_jcp(x, n, vnow, dvdx, zmax, a)
      end if
      do i=1,n
        accelx(i) = -dvdx(i)/rmass
      end do

      do 200 iouter=1,nouter

         ncross=0
         do 100 istep=1,nstep

c  Brownian integration step using Allen and Tildesley method, p 263.

            do i=1,n
               call bivariate(sigr,sigv,crv, randx(i),randvx(i))
               x(i)=x(i) +c1*dt*velx(i) +c2*dt**2*accelx(i) +randx(i)
            end do

c NOTE - should really modify this to handle the cusps as the flat
c        potential begins and ends.  For now, use a small dt.

            call v(x,n, vnow)
            if (zmax .lt. 0.d0) then
               delv = vbias_flat(x, n, vnow, dvdx, vflat)
            else
               delv = vbias_jcp(x, n, vnow, dvdx, zmax, a)
            end if
            do i=1,n
               accelnewx(i) = -dvdx(i)/rmass 
               velx(i) = c0*velx(i) + (c1-c2)*dt*accelx(i) +
     &              c2*dt*accelnewx(i) + randvx(i)
               accelx(i)=accelnewx(i) ! ready for next time
            end do

c  hyper clock accumulation
            boostfac=exp(delv/tk)
            thyper=thyper+dt*boostfac
            tmd=tmd+dt
            if(iprint.ge.2) then ! optional every-step printout
               call v(x,n, vnow)
               write(13,113) x(1),x(2),vnow,vnow+delv,delv,boostfac
 113           format(2f7.3,4f10.5)
            end if


c  state determination

            istate=kstate(x,n)
            if(istate.ne.istatelast) then
               ncross=ncross+1
               ncrosscum=ncrosscum+1
               if(ncrosscum.le.maxtau) then
                  tauobserved(ncrosscum)=thyper-thyperlast
                  iphase=abs(istate-istate2last)-1
                  if(iprint.ge.1) write(11,95) ncrosscum, tmd,
     &                 thyper, thyper-thyperlast, istate, iphase
 95               format(i8,1p3e14.6,2i5)
                  thyperlast=thyper
               end if
               istate2last=istatelast
            end if
            istatelast=istate

 100     continue
      
         crosscount(iouter) = ncross

c  save kinetic energy, just as a sanity check on the thermostat
         ekin=0.0d0
         do i=1,n
            px = velx(i)*rmass
            ekin=ekin + px*px/2.0d0/rmass
         end do
         tkkin = 2.0d0*ekin/float(n)
         temps(iouter)=tkkin

c  report position, time, kinetic temp, etc if desired to fort.9
         if (iprint.ge.1) then
            call v(x,n, vnow)
            if (zmax .lt. 0.d0) then
               delv = vbias_flat(x, n, vnow, dvdx, vflat)
            else
               delv = vbias_jcp(x, n, vnow, dvdx, zmax, a)
            end if
            tplusv = ekin + vnow + delv

            write(9,160) tmd, thyper, tkkin, vnow,
     &           vnow+delv, boostfac, tplusv, x(1), x(2)
 160        format(1p2e14.6,e12.3,4e14.6,2e12.3)
         end if

 200  continue

c
c  end of simulation; print out final statistics
c
      write(6,220) ncrosscum, float(ncrosscum)/thyper
 220  format(' from ',i10, ' crossings, kTST = ',1pd12.3)

      call ave(crosscount,nouter,crossave,sdmean)
      rate=nouter*crossave/thyper
c this is not nec. proper errbar, since thyper has noise also.
      sdmrate=nouter*sdmean/thyper
      write(6,230) rate,sdmrate
  230 format(' using errbars on cross counts, k=',1pd10.3,' +/-',d10.3)

      if(ncrosscum.gt.0) then
        call ave(tauobserved,ncrosscum,tauave,sdmtau)
        write(6,240) tauave,sdmtau
  240 format(' average of tau observations =',1pd10.3,' +/-',d10.3)
      end if

      write(6,250) tmd,thyper,thyper/tmd
  250 format(/' total MD time =',f20.2,'  total hyperMD time =',1pd12.4/
     x ' average boost ratio =',1pd12.4)

      call ave(temps,nouter,tkave,tksdmean)
      varomean2 = (tksdmean*sqrt(float(nouter+1)))**2/tkave**2
      write(6,270) tkave,tksdmean,varomean2, 2.0d0/float(n)
  270 format(' average kT =',1pd10.3,' +/-',d10.3,/
     x       '  var/mean**2 =',1pd10.3,'  (ideal =',0pf8.3,')')

      stop
      end
